'''
Created on 22 jun 2009

@author: Niels
'''
import pyaudio,audioop
import wave
import sys
import time
import math
from PyQt4 import QtCore, QtGui

from threading import Thread, Timer


class SpeechDetector(Thread):
    
    def __init__(self):
        Thread.__init__(self)
        self.chunk = 4410
        self.FORMAT = pyaudio.paInt16
        self.CHANNELS = 1
        self.RATE = 44100
        self.stop = False
        
        self.threshold = 5000
        self.profiles = []
        self.timeOfSwitch = time.time()
        self.startPrevActivity =  time.time()
        self.startPrevSilence = time.time()
        self.durationLastActivity = 0
        self.durationLastSilence = 0
        self.spoken = False
        self.switch = False
        self.thresholdList = []
        self.detectThresholdOn = False
        self.soundListers = []
        self.start()
    def run(self):
        p = pyaudio.PyAudio()
        

        stream = p.open(format = self.FORMAT,channels = self.CHANNELS,rate = self.RATE,input = True,frames_per_buffer = self.chunk)
        
        while not self.stop:
            data = stream.read(self.chunk)
            if not data == None or len(data) == 0:
                peak = audioop.max(data, 2)
                if self.detectThresholdOn:
                    self.thresholdList.append(peak)
                    self.findThreshold()
                else:           
                    if peak > self.threshold:
                        self.timeOfSwitch = time.time()
                        if not self.spoken:
                            self.switch = True
                            self.spoken = True
                        
                        
                    else:
                        self.timeOfSwitch = time.time()
                        if self.spoken:
                            self.switch = True
                            self.spoken = False
                self.reportLevel(peak)    
            #print peak    
            self.handleTick()
    
    def detectThreshold(self,duration,safeMarginPercentage):
        self.thresholdList = []
        self.detectThresholdOn = True
        self.detectThresholdDuration = duration
        self.detectThresholdStart = time.time()
        self.detectThresholdMargin = safeMarginPercentage
    
    def findThreshold(self):
        
        if time.time() - self.detectThresholdStart >= self.detectThresholdDuration:
            self.detectThresholdOn = False
            
            max = 0
            for item in self.thresholdList:
                if item > max:
                    max = item
            
            self.threshold = max + (max * self.detectThresholdMargin)
            print "new threshold: "+str(self.threshold)            
    
    def addProfile(self,speechDetectorProfile):
        self.profiles.append(speechDetectorProfile)
        
    def removeProfile(self,speechDetectorProfile):
        self.profiles.remove(speechDetectorProfile)
    
    def handleTick(self):
        for profile in self.profiles:
            if self.switch:
                profile.switch(self.spoken,self.timeOfSwitch)
            
            profile.tick()
        
    def getThreshold(self):
        return self.threshold
            
    def addSoundListener(self,function):
        self.soundListers.append(function)
        
    def reportLevel(self,peak):
        for listener in self.soundListers:
            try:
                listener(peak)
            except:
                print "failed to call function: "+str(listener)
                
class SpeechDetectorProfile():
    
    def __init__(self,timeBeforeSwitch,functionWhenSwitchDetected):
        self.timeBeforeSwitch = timeBeforeSwitch
        self.functionSWD = functionWhenSwitchDetected
        self.beginSpoken = None
        self.beginSilence = None
        self.speakingState = None
        self.lastStateSpeaking = None
        self.lastSwitchTime = None
        self.speakingHistory = []
        
        
    def switch(self,speaking,time):
        self.lastSwitchTime = time
        self.speakingState = speaking
            
    def tick(self):
        if self.speakingState == None:
            pass
        else:
            timeNow = time.time()
            if self.speakingState:
                if self.beginSpoken == None:
                    self.beginSpoken = self.lastSwitchTime
                    
                #check if prev silence was long enough
                if self.lastSwitchTime - self.beginSpoken >= self.timeBeforeSwitch:
                    if not self.beginSilence == None:
                        duration = self.beginSpoken - self.beginSilence
                        self.setState(False, duration)
                        self.beginSilence = None
                        
                        
               
                #check if beginSpoken time need to be set
                
                
            else:
                if self.beginSilence == None:
                    self.beginSilence = self.lastSwitchTime
                
                #check if prev speaking was long enough
                if self.lastSwitchTime - self.beginSilence >= self.timeBeforeSwitch:
                    if not self.beginSpoken == None:
                        #print "s :"+str(self.beginSilence)
                        #print "p : "+str(self.beginSpoken)
                        
                        duration = self.beginSilence - self.beginSpoken
                        #print "d: "+str(duration)
                        self.setState(True, duration)
                        self.beginSpoken = None
                
                
                #check if beginSilence time need to be set
                    
    
    def setState(self,speaking,duration):
        if self.lastStateSpeaking == speaking:
            #already informed do nothing
            pass
        else:
            self.fireForSwitch(speaking, duration)
            self.lastStateSpeaking = speaking
             
            
    def fireForSwitch(self,speaking,duration):
        try:
            self.functionSWD(speaking,duration)
        except:
            print "failed to call function: "+str(self.functionSWD) 

class SpeechDetectorWidget(QtGui.QWidget):
    
    def __init__(self,parent,speechDetector):
        
        QtGui.QWidget.__init__(self,parent)
        self.speechDetector = speechDetector
        self.activeLevel = 0
        if not self.speechDetector == None:
            self.speechDetector.addSoundListener(self.updateLevel)
    
    def updateLevel(self,peak):
        self.activeLevel = peak
        self.update()    
    
    def paintEvent(self,event):
        size = self.size()
        #print "draw: "+str(size)
        #check on minimal size (width 5, height 5)
        if size.height() >= 25 and size.width() >= 7:
            #paint
            painter = QtGui.QPainter()
            painter.begin(self)
            
            width = size.width()-2
            height = size.height()-2
            
            painter.drawRect(1,1,width,height)
            
            #draw audio level indicators
            availableHeight = size.height() - 9
            
            levelHeight = availableHeight / 7.0
            levelWidth = size.width() - 6
            levelValues = []
            levelValues.append(self.speechDetector.getThreshold())
            for i in range(1,6):
                levelValues.append(levelValues[0]*(i+1))
            
            counter = 0
            
            while counter < len(levelValues) and levelValues[counter] <= self.activeLevel:
                posX = 3
                posY = height - (counter * 3 + (counter + 1) * levelHeight)
                color = QtGui.QColor(50,205,50)
                if counter >= 3 and counter < 5:
                    color = QtGui.QColor(238,238,0)
                elif counter >= 5:
                    color = QtGui.QColor(255,36,0)
                
                painter.fillRect(posX,posY,levelWidth,levelHeight,color)
                counter += 1
            
            painter.end()
            
class testWindow(QtGui.QMainWindow):
    
    def __init__(self):
        QtGui.QMainWindow.__init__(self)
        SD = SpeechDetector()
        SD.detectThreshold(10,0.2)

        prof = SpeechDetectorProfile(0.1,printMessage)
        SD.addProfile(prof)
        
        self.soundLevelWidget = SpeechDetectorWidget(self,SD)
        self.soundLevelWidget.resize(30,200)
        
        
        self.setCentralWidget(self.soundLevelWidget)
        
 
    def printMessage(spoken,dur):
        print str(spoken)+": "+str(dur) 

    def printLevel(peak):
        print "peak: "+str(peak)
    
def printMessage(spoken,dur):
        print str(spoken)+": "+str(dur)

def printLevel(peak):
        print "peak: "+str(peak)

if __name__ == '__main__':
                                
    app = QtGui.QApplication(sys.argv)
    main = testWindow()
    main.show()
    sys.exit(app.exec_())




        
        